#include "stdio.h"
#include "stdlib.h"
#include "string.h"


////////////////////////////////////////////////////////////

char *header1=NULL, *header2=NULL;

int length1=0, length2=0;
int nAlignments=0;

typedef struct {
    int score;
    int nBlocks;
    int strand;
    int from1,from2,to1,to2;
    int *data;
} alignBlock;

alignBlock *ALT;

int strand1=0, strand2=0;

////////////////////////////////////////////////////////////

void readHeaders(FILE *INPUT);
void readStrands(FILE *INPUT);
void readAlignmentBlock(FILE *INPUT);
void printOutBinary(void);

////////////////////////////////////////////////////////////

int main (int argc, char *argv[]) {
    FILE *INPUT;
    
    if (argc!=2) {
	fprintf(stderr,"Usage: %s blastz_output\n",argv[0]); exit(1); }

    // open input file
    if ((INPUT=fopen(argv[1],"r")) == NULL) {
	fprintf(stderr,"Error. Unable to open %s file\n",argv[1]); exit(1); }

    // read input file and construct binary alignment
    char readStr[255];
    while (fgets(readStr,255,INPUT) != NULL) {

	// headers
	if ( (header1==NULL) && (strstr(readStr,"h {") != NULL) )
	    readHeaders(INPUT);

	// strands (0:forward, 1:rc)
	if ( strstr(readStr,"s {") != NULL )  
	    readStrands(INPUT);

	// alignment blocks
	if ( strstr(readStr,"a {") != NULL )  
	    readAlignmentBlock(INPUT);
    }

    // close input file
    fclose(INPUT);

    // print out the alignment in binary (zbn) format
    printOutBinary();
}

/////////////////////////////////// SUBS ///////////////////////////////

// read first and second sequence header
void readHeaders(FILE *INPUT) {
    char readStr[255];
    if (fgets(readStr,255,INPUT) != NULL) {
	strtok(readStr,"\"");
	header1 = strdup(strtok(NULL,"\""));
    }
    if (fgets(readStr,255,INPUT) != NULL) {
	strtok(readStr,"\"");
	header2 = strdup(strtok(NULL,"\""));
    }
}

/////////////////////////////////////////////////////////////////////

// read strands
void readStrands(FILE *INPUT) {
    char readStr[255];
    if (fgets(readStr,255,INPUT) != NULL)
	sscanf(readStr,"%*s %*s %d %d",&length1,&strand1);
    if (fgets(readStr,255,INPUT) != NULL)
	sscanf(readStr,"%*s %*s %d %d",&length2,&strand2);
    //printf("strands:%d %d\n",strand1,strand2);
}

/////////////////////////////////////////////////////////////////////

// read alignment block
void readAlignmentBlock(FILE *INPUT) {
    char readStr[255];
    int s1, b1,b2, e1,e2, l1,l2,l3,l4,l5;
    
    ++nAlignments;
    if ((ALT = (alignBlock *) realloc(ALT,sizeof(alignBlock)*nAlignments)) == NULL) {
	printf ("Error. Lack of memory...\n"); exit(1); }

    int i=nAlignments-1;
    ALT[i].nBlocks=0;
    if (strand1==strand2) { ALT[i].strand=0; } else { ALT[i].strand=1; }
    while ((fgets(readStr,255,INPUT) != NULL) && (strstr(readStr,"}")==NULL)) {
        if (sscanf(readStr," %*[l] %d %d %d %d %d",&l1,&l2,&l3,&l4,&l5)==5) {
	    ++ ALT[i].nBlocks;
	    if (ALT[i].nBlocks==1) ALT[i].data=NULL;
    	    if ((ALT[i].data = (int *) realloc(ALT[i].data,5*sizeof(int)*ALT[i].nBlocks)) == NULL) {
		printf ("Error. Lack of memory...\n"); exit(1); }

	    //from1,from2,to1,to2,percent
	    int j = 5* (ALT[i].nBlocks-1);
	    ALT[i].data[j+0] = l1;
	    ALT[i].data[j+1] = l2;
	    ALT[i].data[j+2] = l3;
	    ALT[i].data[j+3] = l4;
	    ALT[i].data[j+4] = l5;
        } 
	else if (sscanf(readStr," %*[b] %d %d", &ALT[i].from1, &ALT[i].from2)==2) {}
        else if (sscanf(readStr," %*[e] %d %d", &ALT[i].to1, &ALT[i].to2)==2) {}
        else if (sscanf(readStr," %*[s] %d", &ALT[i].score)==1) {}
        else return;
    }

    return;
}

/////////////////////////////////////////////////////////////////////

// print out the alignment in binary (zbn) format
void printOutBinary(void) {

   // 0. intergrity check
   if ( (header1==NULL) || (header2==NULL) || (length1<=0) || (length2<=0) || (nAlignments<=0) || (ALT[0].data==NULL) ) {
	fprintf(stderr,"Error. Unsupported blastz output format...\n");
	exit(1);
   }

   // 1. single character output -> size of INT in bytes
   char sizeOfInt;
   sizeOfInt = sizeof(int);
   fwrite (&sizeOfInt,1,1,stdout);
	
   // 2. basic headers
   int header1len, header2len;
   header1len = strlen(header1);
   header2len = strlen(header2);
   fwrite (&header1len, sizeof(int), 1, stdout);
   fwrite (header1, header1len, 1, stdout);
   fwrite (&header2len, sizeof(int), 1, stdout);
   fwrite (header2, header2len, 1, stdout);
   fwrite (&length1, sizeof(int), 1, stdout);
   fwrite (&length2, sizeof(int), 1, stdout);

   // 3. alignments
   fwrite (&nAlignments, sizeof(int), 1, stdout);
   int i,j,k;
   for (i=0; i<nAlignments; i++) {
   	fwrite (&ALT[i].score, sizeof(int), 1, stdout);
   	fwrite (&ALT[i].strand, sizeof(int), 1, stdout);
   	fwrite (&ALT[i].from1, sizeof(int), 1, stdout);
   	fwrite (&ALT[i].from2, sizeof(int), 1, stdout);
   	fwrite (&ALT[i].to1, sizeof(int), 1, stdout);
   	fwrite (&ALT[i].to2, sizeof(int), 1, stdout);
   	fwrite (&ALT[i].nBlocks, sizeof(int), 1, stdout);
	for (j=0; j<ALT[i].nBlocks; j++) 
	    for (k=0; k<5; k++) 
   		fwrite (&ALT[i].data[j*5+k], sizeof(int), 1, stdout);
   }
}

/* Binary format
1.
   <size of int>
2.
   <header1 length> <header1 string> <header2 length> <header2 string>
   <sequence1 length> <sequence2 length>
3.
   <nAlignments>
	alignment1: <score> <strand> <from1> <from2> <to1> <to2> <nBlocks>
			block1: <from1> <from2> <to1> <to2> <percent id>
			   ...
			blockN: <from1> <from2> <to1> <to2> <percent id>
	   ...
        alignmentN: ...

   All the records are INTs (int'egers), except for the header strings that are ... strings.
*/
/////////////////////////////////////////////////////////////////////
