#include "draw.h"

time_t timeIni=time(NULL);

//////////////////////////////////////////////////////////////////////////////////////////////

void CreatePicture(long int from, long int to) {
    gdImagePtr im;
    gdFontPtr myFont = gdFontTiny;
    int iLayer, nLayers;
    char sPos[15];

    // define an image
    nLayers = 1 + (to-from+1 - 1)/nLayerBases;
    im = gdImageCreateTrueColor(xMax,nLayers*yLayer+9); 

    // allocate some colors -- white first to make it as a background color
    yellishBkgnd = gdImageColorAllocate(im,255,255,240);
    gdImageFilledRectangle(im,0,0,xMax-1,nLayers*yLayer+9-1, yellishBkgnd);
    white = gdImageColorAllocate(im,255,255,255); 
    gray = gdImageColorAllocate(im,205,205,205); 
    black = gdImageColorAllocate(im,0,0,0); 
    blue = gdImageColorAllocate(im,0,0,255); 
    yellow = gdImageColorAllocate(im,255,255,240);
    red = gdImageColorAllocate(im,255,0,0);
    green = gdImageColorAllocate(im,0,200,0);
    
    // 0 - ncECR, 1 - Intron, 2 - Exon, 3 - UTR, 4 - repeat
    color[0] = gdImageColorAllocate(im,255,50,50);
    color[1] = gdImageColorAllocate(im,255,185,185);
    color[2] = gdImageColorAllocate(im,0,0,255);
    color[3] = gdImageColorAllocate(im,245,245,0);
    color[4] = gdImageColorAllocate(im,25,225,25);
    color[5] = gdImageColorAllocate(im,0,0,0);
    // OTH features
    color[6] = gdImageColorAllocate(im,150,0,150);
    color[7] = gdImageColorAllocate(im,0,0,0);
    
    // Plot every Layer
    for (iLayer=0; iLayer<nLayers; iLayer++) {
	// define vertical boundaries of the layer
	int yLFrom = int(1.0*iLayer*yLayer) + 14+4;
	int yLTo = int(1.0*(iLayer+1)*yLayer-1.0)+4;
	int posFrom = from + int(1.0*iLayer*nLayerBases);
	int posTo = from + int(1.0*(iLayer+1)*nLayerBases-1);
	posTo = (posTo>to) ? to : posTo;
	int xFrom = xPICFROM;
	int xTo = xMax;
	if (iLayer==(nLayers-1))
	    xTo = xMax-(xMax-xPICFROM-xRIGHTSHIFT)*(nLayers*nLayerBases - (to-from+1))/nLayerBases;

	if ((iLayer==nLayers-1) && ((xTo-xFrom+1)<xRIGHTSHIFT)) xRIGHTSHIFT=0; 
	if (xTo<xFrom+1) xTo=xFrom+1;

	// Create the CONSERVATION GRAPH(s) and the ECR TRACK(s)
	int nGraphs = numSeq - 2 + 1;
	for (int iG=0; iG<nGraphs; iG++) {
	    int yPlotFrom = yLFrom+ANNO_HEIGHT+4;
	    int yPlotTo = yLTo-10;
	    int yPlotWidth = yPlotTo - yPlotFrom + 1;
	    int yGFrom = yPlotFrom + (yPlotWidth * iG)/nGraphs;
	    int yGTo = (iG == nGraphs-1) ? yPlotTo : yPlotFrom + (yPlotWidth * (iG+1))/nGraphs - 1;
	    PlotGraph(im,posFrom,posTo,xFrom,xTo-xRIGHTSHIFT,
		      yGFrom, yGTo, iG+2);
	    gdImageLine(im,xFrom,yGTo,xTo-xRIGHTSHIFT,yGTo, black);
	}

	// plot annotations 
	PlotAnno(im,posFrom,posTo,xFrom,xTo-xRIGHTSHIFT,yLFrom,yLFrom+ANNO_HEIGHT-1, yLTo);

	// show axis
	gdImageLine(im,xFrom,yLTo-10,xTo-xRIGHTSHIFT,yLTo-10, black);
	for (int j=0; j<9; j++) { 
		int tickX = xFrom + j*(xTo-xRIGHTSHIFT-xFrom)/8; 
		gdImageLine(im, tickX,yLTo-10, tickX,yLTo-7, black);
	}
	for (int j=0; j<5; j++) { 
		int tickX = xFrom + j*(xTo-xRIGHTSHIFT-xFrom)/4; 
		gdImageLine(im, tickX,yLTo-10, tickX,yLTo-5, black);
		num2str(sPos, posFrom-from + j*(posTo-posFrom+1)/4);
		if (((xTo-xFrom)>400) || (j==0) || (j==4)) 
			gdImageStringFT (im, NULL, black, fontComic, 8., 0.0, tickX-8*j, yLTo+5, sPos);
	}
    }

    // print the PNG image to the stdout
    gdImagePng(im,stdout);
    
    gdImageDestroy(im);
}

//////////////////////////////////////////////////////////////////////////////////////////////

void num2str(char *str, int num) {
	if (num>=1000) {
		sprintf(str, "%.1fkb", 1.0*num/1000.0);
	} else {
		sprintf(str, "%d", num);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////

void PlotGraph(gdImagePtr im, long int from, long int to, int xFrom, int xTo, int yFrom, int yTo, int iS) {
    gdFontPtr myFont = gdFontTiny;
    gdFontPtr myMediumFont = gdFontMediumBold;
    
    // shift yFrom
    yFrom += ANNO_HEIGHT;

    // white background
    gdImageFilledRectangle(im,xFrom,yFrom,xTo,yTo, white);
   
    // sequence titles
    char seqTit1[255], seqTit2[255];
    if ((header1[iS] != NULL) && (header2[iS] != NULL)) {
	if (topBottom) {
	    	strncpy(seqTit1, header1[iS]+1, 12);
    		strncpy(seqTit2, header2[iS]+1, 12);
	} else {
	    	strncpy(seqTit2, header1[iS]+1, 12);
    		strncpy(seqTit1, header2[iS]+1, 12);
	}
    	seqTit1[12] = seqTit2[12] = 0;
	gdImageStringFT (im, NULL, black, fontComicBold, 8., 0.0, xTo+10, (yFrom+yTo)/2+13, seqTit2);
    	gdImageStringFT (im, NULL, black, fontComic, 7., 0.0, xTo+15, (yFrom+yTo)/2+3, "- vs -");
	gdImageStringFT (im, NULL, black, fontComicBold, 8., 0.0, xTo+10, (yFrom+yTo)/2-7, seqTit1); 
    }

    // colorize the background of alignments on the reverse strand
    for (int iA=0; iA<nAlignments[iS]; iA++) {
	int aPosFrom, aPosTo;
	int strand = ALT[iS][iA].strand;
	if ((strand==1) && topBottom) { // negative strand - second sequence
	    aPosFrom = length2[iS] - ALT[iS][iA].to2 + 1;
	    aPosTo = length2[iS] - ALT[iS][iA].from2 + 1;
	} else {
	    aPosFrom = ALT[iS][iA].from1;
	    aPosTo = ALT[iS][iA].to1;
	}
    	int altColor = gdImageColorAllocate(im,210,210,210);
	if ((aPosFrom <= to) && (from <= aPosTo) && (strand==1)) {
	    aPosFrom = (aPosFrom>from) ? aPosFrom : from;
	    aPosTo = (aPosTo<to) ? aPosTo : to;
	    int xx1 = xFrom + int(0.5+1.0*(xTo-xFrom)*(aPosFrom-from)/(to-from));
	    int xx2 = xFrom + int(0.5+1.0*(xTo-xFrom)*(aPosTo-from)/(to-from));
	    gdImageFilledRectangle(im,xx1,yFrom,xx2,yTo, altColor);
	}
    }
 
    // top & middle lines
    gdImageLine(im,xFrom,yFrom,xTo,yFrom, black);
    gdImageLine(im,xFrom,(yFrom+yTo)/2,xTo,(yFrom+yTo)/2, gray);

    // draw ECR bars
    int cnsYFrom = yFrom - 2;
    int cnsYTo = yFrom + 2;
    for (int iE=0; iE<nECR[iS]; iE++) {
	long int cPosFrom = ECR[iS][iE*3 + 0];
	long int cPosTo = ECR[iS][iE*3 + 1];
    	int ecrColor = gdImageColorAllocate(im,155,50,50);
	if ((cPosFrom <= to) && (from <= cPosTo)) {
	    long int type = ECR[iS][iE*3 + 2];
	    cPosFrom = (cPosFrom>from) ? cPosFrom : from;
	    cPosTo = (cPosTo<to) ? cPosTo : to;
	    int xx1 = xFrom + int(0.5+1.0*(xTo-xFrom)*(cPosFrom-from)/(to-from));
	    int xx2 = xFrom + int(0.5+1.0*(xTo-xFrom)*(cPosTo-from)/(to-from));
	    gdImageFilledRectangle(im,xx1,cnsYFrom,xx2,cnsYTo, ecrColor);
	}
    }
 
    // construct a graph
    int yPrev=0, xPrev=-1;
    /* plot the graph */
    int x;
    for (x=xFrom; x<=xTo; x++) {
	unsigned long int xPos = from + int (0.5+1.0*(to-from)*(x-xFrom)/(xTo-xFrom));
	unsigned long int shift = xPos;
	int yGraph;
	if (!SMOOTH) {
	    // PIP plot
	    yGraph = PipConservation[iS][xPos];
	} 
	else {
	    // SMOOTH plot
	    yGraph = Conservation[iS][xPos];
	}
	int y = int(1.0 * min(max(yGraph-percFrom,0),(percTo-percFrom)) * (yTo-yFrom) / (percTo-percFrom));

	// Plot the ECR line
        bool insideGene=false;
        for (int iGene=0; iGene<nGenes; iGene++) {
	    unsigned long int gFrom = Gene[iGene].from;
	    unsigned long int gTo = Gene[iGene].to;
	    if ((gFrom <= xPos) && (xPos <= gTo)) 
		insideGene=true;
	}
	for (int iE=0; iE<nECR[iS]; iE++) {
	    unsigned long int cPosFrom = ECR[iS][iE*3 + 0];
	    unsigned long int cPosTo = ECR[iS][iE*3 + 1];
	    int type = ECR[iS][iE*3 + 2];
	    if (insideGene && type==0) { type=1; }
	    if ((cPosFrom <= xPos) && (xPos <= cPosTo)) 
		gdImageLine(im,x,yTo,x,yTo-y,color[type]);
	}

        // draw CODING, UTR or REPEAT line
	for (int iAnno=0; iAnno<nAnnoLines; iAnno++) {
	    long int fFrom = Annotation[iAnno*3+0];
	    long int fTo = Annotation[iAnno*3+1];
	    long int fType = Annotation[iAnno*3+2];
	    if ((fFrom <= xPos) && (xPos <= fTo) && (fType>=2) && (fType<=6)) 
		gdImageLine(im,x,yTo,x,yTo-y,color[fType]);
	}
	
	// Plot the black graph
	if (SMOOTH) {
		if ((x==xTo) || ((x-xPrev)>1))
		    // First point or the region starting position
		    gdImageLine(im,x,yTo-y,x,yTo,black);
		else
		    // Normal line from the first to the last point
		    gdImageLine(im,x-1,yTo-yPrev,x,yTo-y,black);
		// Plotting the final point
		if (x==xTo)
		    gdImageLine(im,x,yTo-y,x,yTo,black);
		yPrev=y; 
		xPrev=x;
	}
    }

    // draw PIPs
    if (!SMOOTH) {
    	// Choose sequence to build an annotation on top of
    	int seqInd = 1;
    	if (topBottom) { seqInd=2; }

    	// Plot PIP lines
    	int i,j;
    	for (i=0; i<nAlignments[iS]; i++) {
	    int strand = ALT[iS][i].strand; // 0,1 (positive,negative)
	    for (j=0; j<ALT[iS][i].nBlocks; j++) {
	    	int f[3], t[3];
		int plotFrom,plotTo;
		unsigned long int xPosFrom, xPosTo;
	    	f[1] = ALT[iS][i].data[j*5+0];
	    	f[2] = ALT[iS][i].data[j*5+1];
	    	t[1] = ALT[iS][i].data[j*5+2];
	    	t[2] = ALT[iS][i].data[j*5+3];
	    	int pID = ALT[iS][i].data[j*5+4];
	    	if ((strand==0) || (seqInd==1)) {
		    plotFrom = f[seqInd];
		    plotTo = t[seqInd];
	    	} else {
		    plotFrom = length2[iS]-t[seqInd]+1;
		    plotTo = length2[iS]-f[seqInd]+1;
		}
		xPosFrom = xFrom + int (0.5+1.0*(xTo-xFrom)*(plotFrom-from)/(to-from));
		xPosTo = xFrom + int (0.5+1.0*(xTo-xFrom)*(plotTo-from)/(to-from));
		if ((xFrom<=xPosTo) && (xPosFrom<=xTo)) {
		    if (xPosTo>xTo) { xPosTo=xTo; }
		    if (xPosFrom<xFrom) { xPosFrom=xFrom; }
		    int y = int(1.0 * min(max(pID-percFrom,0),(percTo-percFrom)) * (yTo-yFrom) / (percTo-percFrom));
		    gdImageLine(im,xPosFrom,yTo-y,xPosTo,yTo-y,black);
		    if (y>0)
		        gdImageLine(im,xPosFrom,yTo-y+1,xPosTo,yTo-y+1,black);
		}
	    }
	}
    }

    // Plot the legend
    int type;
    char sPerc[10];
    char lName[5][11] = {"intergene","intron","coding","UTR","repeat"};
    int barWidth = 10;
    if ((!removeLegend) && (iS==2)) {
	for (type=0; type<5; type++) {
	    gdImageFilledRectangle(im,x1+2,yFrom+20+(type-1)*barWidth,x1+2+barWidth,yFrom+20+type*barWidth-3,color[type]);
	    gdImageStringFT (im, NULL, black, fontComic, 7., 0.0, x1+4+barWidth, yFrom+20+type*barWidth-4, lName[type]);
	}
	sprintf(sPerc,"%d%%",percTo);
	gdImageStringFT (im, NULL, black, fontComic, 8., 0.0, xTo+3, yFrom+6, sPerc);
	sprintf(sPerc,"%d%%",percFrom);
	gdImageStringFT (im, NULL, black, fontComic, 8., 0.0, xTo+3, yTo+4, sPerc);
    }
    
}
    
/////////////////////////////////////////////////////////////////////////////////////////////

void PlotAnno(gdImagePtr im, long int from, long int to, int xFrom, int xTo, int yFrom, int yTo, int yLTo) {
    gdFontPtr myFont = gdFontTiny;
    int nPrintedGenes(0);
    
    // draw Genes
    int yGeneFrom = yFrom+ANNO_HEIGHT+1;
    int yGeneTo = yTo+ANNO_HEIGHT;
    for (int iGene=0; iGene<nGenes; iGene++) {
	unsigned long int gFrom = Gene[iGene].from;
	unsigned long int gTo = Gene[iGene].to;
	if ((gFrom <= to) && (from <= gTo)) {
	    gFrom = (gFrom>from) ? gFrom : from;
	    gTo = (gTo<to) ? gTo : to;
	    int xx1 = xFrom + int(0.5+1.0*(xTo-xFrom+1)*(gFrom-from+1)/(to-from+1));
	    int xx2 = xFrom + int(0.5+1.0*(xTo-xFrom+1)*(gTo-from+1)/(to-from+1));
	    // draw the line
	    gdImageLine(im,xx1,int((yGeneFrom+yGeneTo)/2.0),xx2,int((yGeneFrom+yGeneTo)/2.0),color[2]);
	    // draw the directions
	    int xDir;
	    for (xDir=xx1; xDir<=max(xx2-3,xx1); xDir += 8) {
		if (Gene[iGene].strand==1) {
		    gdImageLine(im,xDir+3,int((yGeneFrom+yGeneTo)/2.0),xDir,int(yGeneFrom+1),color[2]);
		    gdImageLine(im,xDir+3,int((yGeneFrom+yGeneTo)/2.0),xDir,int(yGeneTo-1),color[2]);
		} else {
		    gdImageLine(im,xDir,int((yGeneFrom+yGeneTo)/2.0),xDir+3,int(yGeneFrom+1),color[2]);
		    gdImageLine(im,xDir,int((yGeneFrom+yGeneTo)/2.0),xDir+3,int(yGeneTo-1),color[2]);
		}
	    }
	    // print out the gene name
	    int xGeneStrFrom =  (xx1+xx2)/2-8*strlen(Gene[iGene].name)/2;
	    if (xGeneStrFrom<0) xGeneStrFrom=0;
	    gdImageStringFT (im, NULL, black, fontTimes, 12., 0.0, xGeneStrFrom,  (yFrom+yTo)/2+4,  Gene[iGene].name);
	}
    }

    // draw CODING, UTR & REPEAT BARS
    for (int iAnno=0; iAnno<nAnnoLines; iAnno++) {
	long int fFrom = Annotation[iAnno*3+0];
	long int fTo = Annotation[iAnno*3+1];
	long int fType = Annotation[iAnno*3+2];
	int shift = 0;
	if (fType==4) shift=1;
	if ((fFrom <= to) && (from <= fTo) && (fType>=2) && (fType<=6)) {
	    fFrom = (fFrom>from) ? fFrom : from;
	    fTo = (fTo<to) ? fTo : to;
	    int xx1 = xFrom + int(0.5+1.0*(xTo-xFrom)*(fFrom-from)/(to-from));
	    int xx2 = xFrom + int(0.5+1.0*(xTo-xFrom)*(fTo-from)/(to-from));
	    // draw the line
	    gdImageFilledRectangle(im,xx1,yGeneFrom+shift*(yLTo-12-yGeneFrom),xx2,yGeneTo+shift*(yLTo-yGeneTo-8),color[fType]);
	}
    }
}
//////////////////////////////////////////////////////////////////////////////////////////////


