#include "ecr.h"

//////////////////////////////////////////////////////////////////////////////////////////////

void Define_ECRs( int iS ) {

    // Init/NULL ECR array
    nECR[iS] = 0;
    ECR[iS] = NULL;

    // Define base sequence index
    int seqInd = 1;
    if (topBottom) { seqInd = iS; }

    // Define ECRes
    long int ecrFrom(0), ecrTo(0);
    //// step 1. number of conserved bps in the initial left-most window
    int sum=0;
    for (int i=0; i<ecrWin; i++) 
	if (isupper( seq[seqInd][i] ))
	    ++sum;
    if (sum>=int(0.5+ecrId*ecrWin/100.0)) {
	ecrTo = ecrWin;
	ecrFrom = 1;
    }
    //// step 2. scan the match-mismatch profile
    int seqLen = strlen( seq[seqInd] );
    for (int i=0; i<seqLen-ecrWin; i++) {
	if (isupper(seq[seqInd][i])) 
	    --sum;
	if (isupper(seq[seqInd][i+ecrWin])) 
	    ++sum;

	if (sum>=int(0.5+ecrId*ecrWin/100.0)) {
	    if (ecrFrom==0) 
	        ecrFrom = i+2;
	    ecrTo = i+ecrWin+1;
	} else {
	    if (ecrFrom>0) 
		storeECR(ecrFrom,ecrTo, iS, seqInd);
	    ecrFrom = ecrTo = 0;
	}
    }
    if (ecrFrom>0) 
	storeECR(ecrFrom,ecrTo, iS, seqInd);
}


//////////////////////////////////////////////////////////////////////////////////////////////

void storeECR(long int ecrFrom, long int ecrTo, int iS, int seqInd) {
    // trim boundaries to a match
    while (! isupper( seq[seqInd][ecrFrom-1] )) ++ecrFrom;
    while (! isupper( seq[seqInd][ecrTo-1] )) --ecrTo;
	
    // Store an ECR
    ++nECR[iS];
    if ((ECR[iS] = (long int *) realloc(ECR[iS], 3*nECR[iS]*sizeof(long int))) == NULL) { printf("Error: Not enough memory for ECR data storage (realloc)\n"); exit(1); }
    
    ECR[iS][3*(nECR[iS]-1)+0] = ecrFrom;
    ECR[iS][3*(nECR[iS]-1)+1] = ecrTo;
    ECR[iS][3*(nECR[iS]-1)+2] = 0; // intergene, no type definition at this point
}

//////////////////////////////////////////////////////////////////////////////////////////////

void JoinECROverlaps(int iS) {
    int i=0;
    while (i<nECR[iS]-1) {
	int ecrFrom1 = ECR[iS][3*(i)+0];
	int ecrTo1 = ECR[iS][3*(i)+1];
	int ecrFrom2 = ECR[iS][3*(i+1)+0];
	int ecrTo2 = ECR[iS][3*(i+1)+1];
	if ((ecrFrom1<=ecrTo2) && (ecrFrom2<=ecrTo1)) {
		ECR[iS][3*(i)+0] = min(ecrFrom1,ecrFrom2);
		ECR[iS][3*(i)+1] = max(ecrTo1,ecrTo2);
		for (int j=i+1; j<nECR[iS]-1; j++) {
		    ECR[iS][3*(j)+0] = ECR[iS][3*(j+1)+0];
		    ECR[iS][3*(j)+1] = ECR[iS][3*(j+1)+1];
		}
		--nECR[iS];
	} else 
		++i;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////

void PrintEcrs(long int from, long int to, int iS) {
    // Define base sequence index
    int seqInd = 1;
    if (topBottom) { seqInd = iS; }

    // Scan through ECRs
    bool headerPrinted = false;
    for (int iE=0; iE<nECR[iS]; iE++) {
	int ecrFrom = ECR[iS][3*iE+0];
	int ecrTo = ECR[iS][3*iE+1];
	int ecrLen = ecrTo-ecrFrom+1;
	int pID;
	
	if ((ecrFrom<=to) && (from<=ecrTo)) {
	    // print header
	    if (!headerPrinted) {
		printf("Position\tLength\tPercent identity\n");
		headerPrinted=true; }
	    
	    // calculate percent identity
	    int sum=0;
	    for (int i=ecrFrom-1; i<ecrTo; i++) 
		if (isupper( seq[seqInd][i] ))
		    ++sum;
	    pID = int(0.5 + 100.0*sum/ecrLen);
	    
	    // output an ECR
	    printf("%d-%d\t%dbp\t%d%%",ecrFrom,ecrTo,ecrLen,pID);

	    // overlap with an annotation feature?
	    //// 1. gene overlap?
	    char *gene=NULL;
	    bool insideGene=false;
	    for (int iGene=0; iGene<nGenes; iGene++) {
		unsigned long int gFrom = Gene[iGene].from;
		unsigned long int gTo = Gene[iGene].to;
		if ((gFrom <= ecrTo) && (ecrFrom <= gTo))
		    gene=Gene[iGene].name;
	    }
	    //// 2. CODING, UTR or REPEAT overlap?
	    bool annoHeaderPrinted=false;
	    for (int iAnno=0; iAnno<nAnnoLines; iAnno++) {
		long int fFrom = Annotation[iAnno*3+0];
		long int fTo = Annotation[iAnno*3+1];
		long int fType = Annotation[iAnno*3+2];
		if ((fFrom <= ecrTo) && (ecrFrom <= fTo) && (fType>=2) && (fType<=4)) {
		    char desc[255];
		    switch (fType) {
		    case (2): sprintf(desc,"coding exon of %s",gene); break;
		    case (3): sprintf(desc,"UTR of %s",gene); break;
		    case (4): sprintf(desc,"repeat"); break;
		    }
		    if (!annoHeaderPrinted) {
			printf("\toverlaps with:\n");
			annoHeaderPrinted=true;
		    }
		    printf("\t\t\t\t%d-%d\t%dbp\t%s\n",fFrom,fTo,(fTo-fFrom+1),desc);
		}
	    }
	    printf("\n");
	}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////

