#include "readAlignment.h"

/////////////////////////////////////////////////////////////////////////////////////////////

// Reading sequences from SeqXFileName to global char* seqX array 
void ReadSeqs() {
    FILE *seqFile;
    long int fSize;
    char *fName;

    // read sequences seq1 through seq(numSeq)
    for (int i=1; i<=numSeq; i++) {
	fName = GenSeqFileName[i];

	// open sequence file
        if ((seqFile = fopen(fName,"r")) == NULL) { fprintf(stderr,"Error. Can not open %s file\n",fName); exit(1); }

	// find out file size --> allocate strings
	fseek(seqFile,0L,SEEK_END);
	if ((fSize = ftell(seqFile)) == 0L) { fprintf(stderr,"Error: Zero sized sequence file...\n"); exit(1); }
	rewind(seqFile);
	
	// allocate memory for sequences
	if ((seq[i] = (char *) malloc(sizeof(char)*( fSize+1+1000 ))) == NULL) { fprintf(stderr,"Error. Lack of memory...\n"); exit(1); }
	seq[i][0] = 0;
	
	// read FASTA files
	char buffer[1000];
	int seqLen=0;
    	while ((seqFile!=NULL) && (fgets(buffer,998,seqFile) != NULL)) 
	    if (index(buffer,'>')==NULL) {
                int bufLen=strlen(buffer);
		if (buffer[bufLen-1]=='\r') { buffer[bufLen-1]=0; }
		if (buffer[bufLen-1]=='\n') { buffer[bufLen-1]=0; }
		if (buffer[bufLen-2]=='\r') { buffer[bufLen-2]=0; }
		if (buffer[bufLen-2]=='\n') { buffer[bufLen-2]=0; }
		bufLen=strlen(buffer);
		memcpy(seq[i]+seqLen, buffer, bufLen);
		seqLen+=bufLen;
		seq[i][seqLen]=0;
	    }
	
	// close file
	fclose(seqFile);
    }

}

/////////////////////////////////////////////////////////////////////////////////////////////

char RC(char x) {
    switch (x) {
	case 'a': return 't'; break;
	case 'c': return 'g'; break;
	case 't': return 'a'; break;
	case 'g': return 'c'; break;
	case 'A': return 'T'; break;
	case 'C': return 'G'; break;
	case 'T': return 'A'; break;
	case 'G': return 'C'; break;
        default : return x;
    }
}

/////////////////////////////////////////////////////////////////////////////////////////////

void ReformatSeqs( int iS ) {

    // move seq[1] seq[iS] sequences to lower case format initially
    for (int i=1; i<=iS; i+=(iS-1)) {
	int sLen = strlen(seq[i]);
	for (int j=0; j<sLen; j++) 
		seq[i][j] = tolower(seq[i][j]);
    }	     

    // sanity check point
    if (strlen(seq[1]) != length1[iS])
	return;
    if (strlen(seq[iS]) != length2[iS])
	return;

    // process all the alignments ALT[] and move to upper case matching bases
    int i,j;
    for (i=0; i<nAlignments[iS]; i++) {
	// ALT[iS][i].strand -> 0,1 (positive,negative)
	int strand = ALT[iS][i].strand;
	for (j=0; j<ALT[iS][i].nBlocks; j++) {
	    int from[3], to[3];
	    from[1] = ALT[iS][i].data[j*5+0];
	    from[2] = ALT[iS][i].data[j*5+1];
	    to[1] = ALT[iS][i].data[j*5+2];
	    to[2] = ALT[iS][i].data[j*5+3];
	    int pID = ALT[iS][i].data[j*5+4];
	    //printf("i:%d strand:%d 1:%d-%d 2:%d-%d pID:%d\n",i,strand,from[1],to[1],from[2],to[2],pID);
           
	    // move to UPPER case matching bps
 	    int k2;
	    for (int k1=from[1]; k1<=to[1]; k1++) {
		if (strand==0) {
		    k2=k1-from[1]+from[2];
		    if (tolower(seq[1][k1-1]) == tolower(seq[iS][k2-1])) {
			seq[1][k1-1] = toupper(seq[1][k1-1]);
			seq[iS][k2-1] = toupper(seq[iS][k2-1]);
		    }  
		}
		// negative strand
		else {
		    k2=length2[iS] - (k1-from[1]) - from[2] + 1;
		    if (tolower(seq[1][k1-1]) == tolower(RC(seq[iS][k2-1]))) {
			seq[1][k1-1] = toupper(seq[1][k1-1]);
			seq[iS][k2-1] = toupper(seq[iS][k2-1]);
		    }  
		}
	    }

/*
	    for (int s=1; s<=iS; s+=(iS-1)) {
		printf("\t");
	    	for (int k=from[s]; k<=to[s]; k++) {
		    if (s==1 || strand==0) {
		    	printf("%c",seq[s][k-1]);
		    } else {
		    	printf("%c",RC(seq[s][length2[iS]-k]));
		    }
		}
		printf("\n");
	    }
*/

	}
    }

}

/////////////////////////////////////////////////////////////////////////////////////////////

void Get_BLASTZ_Alignment( int iSeq )  {

    FILE *BlastZFile;

    // 0. open file and verify its non-zero size
    unsigned long int posCur;
    if ((BlastZFile = fopen( GenBinAlnFileName[ iSeq ], "r")) == NULL) {
	fprintf(stderr,"Error. Can not open %s file\n", GenBinAlnFileName[ iSeq ]);
	exit(1);
    }

    fseek(BlastZFile,0L,SEEK_END);
    long int nAlignLines;
    if ((nAlignLines = ftell(BlastZFile)) == 0L) {
	header1[iSeq] = header2[iSeq] = NULL;
	length1[iSeq] = length2[iSeq] = 0;
	nAlignments[iSeq] = 0;
	ALT[iSeq] = NULL;
	return;	
	//fprintf(stderr,"Error: Zero sized alignment file...\n");
	//exit(1);
    }
    //printf ("nAlignLines: %d\n",nAlignLines);

/* BLASTZ Binary format
1.
   <size of int>
2.
   <header1 length> <header1 string> <header2 length> <header2 string>
   <sequence1 length> <sequence2 length>
3.
   <nAlignments>
        alignment1: <score> <strand> <from1> <from2> <to1> <to2> <nBlocks>
                        block1: <from1> <from2> <to1> <to2> <percent id>
                           ...
                        blockN: <from1> <from2> <to2> <to2> <percent id>
           ...
        alignmentN: ...

   All the records are INTs (int'egers), except for the header strings that are ... strings.
*/

   rewind(BlastZFile);

   // 1. size of INT in bytes
   char charSizeOfInt;
   int sizeOfInt;
   fread (&charSizeOfInt,1,1,BlastZFile);
   sizeOfInt = charSizeOfInt;
   /* consistency check point */
   if (sizeOfInt<2 || sizeOfInt>8) { fprintf(stderr,"Error. Incorrect blastz binary file format...\n"); exit(1); } 
   //printf("sizeOfInt:%d\n",sizeOfInt);
	
   // 2. basic headers
   int header1len, header2len;
   fread (&header1len,sizeOfInt,1,BlastZFile);
   if ((header1[iSeq] = (char *) malloc(sizeof(char)*( header1len+1 ))) == NULL) {
	fprintf(stderr,"Error. Lack of memory...\n"); exit(1); }
   header1[iSeq][header1len]=0;
   fread (header1[iSeq], header1len,1,BlastZFile);
   fread (&header2len,sizeOfInt,1,BlastZFile);
   if ((header2[iSeq] = (char *) malloc(sizeof(char)*( header2len+1 ))) == NULL) {
	fprintf(stderr,"Error. Lack of memory...\n"); exit(1); }
   header2[iSeq][header2len]=0;
   fread (header2[iSeq], header2len,1,BlastZFile);
   /* consistency check point */
   if (header1len<=0 || header2len<=0) { fprintf(stderr,"Error. Incorrect blastz binary file format...\n"); exit(1); } 
   fread (&length1[iSeq], sizeOfInt, 1, BlastZFile);
   fread (&length2[iSeq], sizeOfInt, 1, BlastZFile);

   // 3. alignments
   fread (&nAlignments[iSeq], sizeOfInt, 1, BlastZFile);
   /* consistency check point */
   if (length1[iSeq]<=0 || length2[iSeq]<=0 || nAlignments[iSeq]<=0) { fprintf(stderr,"Error. Incorrect blastz binary file format...\n"); exit(1); } 
   if ((ALT[iSeq] = (alignBlock *) malloc(sizeof(alignBlock)*( nAlignments[iSeq] ))) == NULL) {
	fprintf(stderr,"Error. Lack of memory...\n"); exit(1); }
   int i,j,k;
   for (i=0; i<nAlignments[iSeq]; i++) {
   	fread (&ALT[iSeq][i].score, sizeOfInt, 1, BlastZFile);
   	fread (&ALT[iSeq][i].strand, sizeOfInt, 1, BlastZFile);
   	fread (&ALT[iSeq][i].from1, sizeOfInt, 1, BlastZFile);
   	fread (&ALT[iSeq][i].from2, sizeOfInt, 1, BlastZFile);
   	fread (&ALT[iSeq][i].to1, sizeOfInt, 1, BlastZFile);
   	fread (&ALT[iSeq][i].to2, sizeOfInt, 1, BlastZFile);
   	fread (&ALT[iSeq][i].nBlocks, sizeOfInt, 1, BlastZFile);
        /* consistency check point */
        if (ALT[iSeq][i].nBlocks<=0) { fprintf(stderr,"Error. Incorrect blastz binary file format...\n"); exit(1); } 
   	if ((ALT[iSeq][i].data = (int *) malloc(sizeof(int)*( ALT[iSeq][i].nBlocks*5 ))) == NULL) {
	    fprintf(stderr,"Error. Lack of memory...\n"); exit(1); }
	for (j=0; j<ALT[iSeq][i].nBlocks; j++) 
	    for (k=0; k<5; k++) 
   		fread (&ALT[iSeq][i].data[j*5+k], sizeOfInt, 1, BlastZFile);
   }
   fclose (BlastZFile);
}
 
/////////////////////////////////////////////////////////////////////////////////////////////

void Get_BLASTZ_Conservation(unsigned int window, int iS) {
    unsigned long int currPos;
    long int currIndex;
    unsigned int half_window = window/2, sum;

    // Get memory for the Conservation array
    int alntLen = length1[iS];
    if (topBottom) { alntLen=length2[iS]; }
    if ((Conservation[iS] = (int *) calloc(consLen, sizeof(int))) == NULL) { fprintf(stderr,"Error: Not enough memory for the conservation array\n"); exit(1); }
   
    // sanity check point
    if (strlen(seq[1]) != length1[iS])
	return;
    if (strlen(seq[iS]) != length2[iS])
	return;
 
    // Choose sequence to build an annotation on top of
    int seqInd = 1;
    if (topBottom) { seqInd = iS; }

    // Calc conservation 
    // RIGHT
    for (sum=0, currPos=alntLen-1; currPos>=alntLen-window; currPos--) {
	if (isupper(seq[seqInd][currPos])) 
   	    ++sum;
	currIndex = currPos + half_window - 1;
	if (currIndex<alntLen)
	    Conservation[iS][currIndex] = int(0.5+100.0*sum/window);
    }
    // LEFT and ini SUM
    for (sum=0, currPos=0; currPos<window; currPos++) {
	if (isupper(seq[seqInd][currPos])) 
   	    ++sum;
	currIndex = currPos - half_window;
	if (currIndex>=0)
	    Conservation[iS][currIndex] = int(0.5+100.0*sum/window);
    }
    // BODY
    for (currPos=0; currPos<alntLen-window; currPos++) {
	if (isupper(seq[seqInd][currPos])) 
	    --sum;
	if (isupper(seq[seqInd][currPos+window])) 
	    ++sum;
	Conservation[iS][currPos+half_window] = int(0.5+100.0*sum/window);
	//printf ("%d\n",currPos+half_window);
    }
}

/////////////////////////////////////////////////////////////////////////////////////////////

void Get_BLASTZ_PipConservation( int iS ) {

    // Get memory for the PipConservation array
    int alntLen = length1[iS];
    if (topBottom) { alntLen=length2[iS]; }
    if ((PipConservation[iS] = (int *) calloc(consLen, sizeof(int))) == NULL) { fprintf(stderr,"Error: Not enough memory for the conservation array\n"); exit(1); }

    // sanity check point
    if (strlen(seq[1]) != length1[iS])
	return;
    if (strlen(seq[iS]) != length2[iS])
	return;
 
    // Choose sequence to build an annotation on top of
    int seqInd = 1;
    if (topBottom) { seqInd = 2; }

    // PIPs become PipConservation lines
    int i,j;
    for (i=0; i<nAlignments[iS]; i++) {
	int strand = ALT[iS][i].strand; // 0,1 (positive,negative)
	for (j=0; j<ALT[iS][i].nBlocks; j++) {
	    int from[3], to[3];
	    from[1] = ALT[iS][i].data[j*5+0];
	    from[2] = ALT[iS][i].data[j*5+1];
	    to[1] = ALT[iS][i].data[j*5+2];
	    to[2] = ALT[iS][i].data[j*5+3];
	    int pID = ALT[iS][i].data[j*5+4];
	    if ((strand==0) || (seqInd==1)) {
		for (int k=from[seqInd]-1; k<to[seqInd]; k++)
		    if (PipConservation[iS][k] < pID)
		        PipConservation[iS][k] = pID;
	    } else {
		for (int k=from[2]-1; k<to[2]; k++)
		    if (PipConservation[iS][length2[iS]-k-1] < pID)
		        PipConservation[iS][length2[iS]-k-1] = pID;
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void calcId( int from, int to ) {
	int sum = 0;
	for (int j=from-1; j<to; j++) 
		if (isupper(seq[1][j])) 
			++sum;
	printf("%.2f%%\n", 100.0*sum/(to-from+1));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
